/*
	event.c

	Generalised event handling for motion

	Copyright Jeroen Vreeken, 2002
	This software is distributed under the GNU Public License Version 2
	see also the file 'COPYING'.

*/

#include "motion.h"
#include "event.h"
#include "video.h"
#include "picture.h"
#include "ffmpeg.h"

/*
 *	Various functions (most doing the actual action)
 */

static void send_sms(char *sms_nr, struct tm *currenttime)
{
	char message[150];

	if (!fork()) {
		syslog(LOG_DEBUG, "Sending SMS message to %s", sms_nr);
		strftime(message, sizeof(message), 
			"Motion detected %Y-%m-%d %T\n", currenttime);
		execlp("sms_client", "sms_client", sms_nr, message, 0);
		syslog(LOG_ERR, "Unable to send SMS message to number %s: %m", sms_nr);
		exit(0);
	}
}

static void send_mail(char *mail_address, struct tm *currenttime)
{
	char message[150];
	int fdpipe[2];

	if (!fork()) {
		syslog(LOG_DEBUG, "Sending mail to %s", mail_address);
		strftime(message, sizeof(message), 
			"This is an automated message generated by motion\n\n"
			"Motion detected: %Y-%m-%d %T\n\n", currenttime);
		pipe(fdpipe);
		close(0);
		dup(fdpipe[0]);
		close(fdpipe[0]);
		write(fdpipe[1], message, strlen(message));
		close(fdpipe[1]);
		execlp("mail", "mail", mail_address, "-s", "MOTION ALERT", 0);
		syslog(LOG_ERR, "Unable to send message to e-mail address %s: %m", mail_address);
		exit(1);
	}
}

/* Execute 'command' with 'arg' as its argument.
   if !arg command is started with no arguments */
static void exec_command (struct context *cnt, char *command, char *arg)
{
	char stamp[PATH_MAX];
	mystrftime(stamp, sizeof(stamp), command, cnt->currenttime, cnt->event_nr, cnt->shots);
	if (arg) {
		strcat(stamp, " ");
		strcat(stamp, arg);
	}

	if (!fork()) {
		vid_close();
		syslog(LOG_DEBUG, "Executing external command '%s'\n", stamp);
		execl("/bin/sh", "sh", "-c", stamp, " &", 0);
		syslog(LOG_ERR, "Unable to start external command '%s' with parameters '%s': %m", stamp, arg);
		exit(1);
	}
}

#ifdef HAVE_MYSQL
void put_mysql (struct config *conf, MYSQL *db, char *filebase, struct tm *currenttime, int type)
{
	char sqlquery[512];
	sprintf(sqlquery, "insert into security set filename='%s', minute=%d, hour=%d, day=%d, month=%d, year=%d, type=%d",
	    filebase,
	    currenttime->tm_min, currenttime->tm_hour,
	    currenttime->tm_mday, currenttime->tm_mon+1,
	    currenttime->tm_year+1900, type);
	if (mysql_query(db, sqlquery)) {
		syslog(LOG_ERR, "Error querying mysql");
		exit(0);
	}
}
#endif

#ifdef HAVE_PGSQL
void put_pgsql (struct config *conf, PGconn *conn, char *filebase, struct tm *currenttime, int type)
{
	char sqlquery[512];
	PGresult *res;

	//use proper sql, and not some strange mysql version :-P
	sprintf(sqlquery, 
	    "insert into security(filename, minute, hour, day, month, year, type) values('%s', '%d', '%d', '%d', '%d', '%d', '%d')",
	    filebase,
	    currenttime->tm_min, currenttime->tm_hour,
	    currenttime->tm_mday, currenttime->tm_mon+1,
	    currenttime->tm_year+1900, type);
	res = PQexec(conn, sqlquery);
	if (PQresultStatus(res) != PGRES_COMMAND_OK)
	{
		syslog(LOG_ERR, "PGSQL query failed: %s", sqlquery);
		PQclear(res);
		exit(1);
	}
}
#endif


/* 
 *	Event handlers
 */

static void event_newfile(int type, struct context *cnt, char *filename, void *mime, struct tm *tm)
{
	printf("File of type %s saved to: %s\n", (char *)mime, filename);
}

static void event_motion(int type, struct context *cnt, char *filename, void *mime, struct tm *tm)
{
	if (!cnt->conf.quiet)
		printf("\a");
//	printf("Motion detected\n");
}

static void onsave_command(int type, struct context *cnt, char *filename, void *mime, struct tm *tm)
{
	if (!strncmp(mime, "image", 5))
		if (cnt->conf.onsavecommand)
			exec_command(cnt, cnt->conf.onsavecommand, filename);
	if (!strncmp(mime, "video", 5))
		if (cnt->conf.onmpegcommand) {
			exec_command(cnt, cnt->conf.onmpegcommand, filename);
	}
}


#ifdef HAVE_MYSQL
static void event_mysql_newfile(int type, struct context *cnt, char *filename, void *mime, struct tm *tm)
{
	int sqltype=0;

	if (!strncmp(mime, "image", 5)) {
		sqltype=1;
		if (filename[strlen(filename)-5]=='m')
			sqltype=4;
		if (!strncmp(filename+strlen(filename)-12, "snapshot", 8))
			sqltype=2;
	}

	if (cnt->conf.mysql_db && sqltype)
		put_mysql(&cnt->conf, cnt->database, filename, tm, sqltype);
	
}
#endif

#ifdef HAVE_PGSQL
static void event_pgsql_newfile(int type, struct context *cnt, char *filename, void *mime, struct tm *tm)
{
	int sqltype=0;

	if (!strncmp(mime, "image", 5)) {
		sqltype=1;
		if (filename[strlen(filename)-5]=='m')
			sqltype=4;
		if (!strncmp(filename+strlen(filename)-12, "snapshot", 8))
			sqltype=2;
	}

	if (cnt->conf.pgsql_db)
		put_pgsql(&cnt->conf, cnt->database_pg, filename, cnt->currenttime, 4);
}
#endif

static void event_firstmotion(int type, struct context *cnt, char *dummy, void *dummy2, struct tm *tm)
{
	if (cnt->conf.sms_nr)
		send_sms(cnt->conf.sms_nr, tm);
	if (cnt->conf.mail_address)
		send_mail(cnt->conf.mail_address, tm);
	if (cnt->conf.externcommand)
		exec_command(cnt, cnt->conf.externcommand, 0);
}

static void event_stop_webcam(int type, struct context *cnt, char *dmy, void *dmy2, struct tm *tm)
{
	webcam_stop(cnt);
}

static void event_webcam_put(int type, struct context *cnt, char *img, void *devpipe, struct tm *tm)
{
	if (cnt->conf.webcam_port) {
		if (type==EVENT_IMAGE_DETECTED && cnt->conf.webcam_motion)
			webcam_put(cnt, img);
		if (type==EVENT_IMAGE && !cnt->conf.webcam_motion)
			webcam_put(cnt, img);
	}
}

static void event_vid_putpipe(int type, struct context *cnt, char *img, void *devpipe, struct tm *tm)
{
	vid_putpipe(*(int *)devpipe, img, cnt->imgs.size);
}

static char *imageext(struct context *cnt)
{
	if (cnt->conf.ppm)
		return "ppm";
	return "jpg";
}

static void event_image_detect(int type, struct context *cnt, char *newimg, void *dummy2, struct tm *tm)
{
	struct config *conf=&cnt->conf;
	char fullfilename[PATH_MAX];
	char filename[PATH_MAX];
	char fullfilenamem[PATH_MAX];
	char filenamem[PATH_MAX];

	if (conf->motion_img || conf->new_img) {
		char *jpegpath;

		/* conf.jpegpath would normally be defined but if someone deleted it my control interface
		   it is better to revert to the default than fail */
		if (cnt->conf.jpegpath)
			jpegpath = cnt->conf.jpegpath;
		else
			jpegpath = DEF_JPEGPATH;
			
		mystrftime(filename, sizeof(filename), jpegpath, cnt->currenttime, cnt->event_nr, cnt->shots);
		/* motion images gets same name as normal images plus an appended 'm' */
		sprintf(filenamem, "%sm", filename);
		sprintf(fullfilename, "%s/%s.%s", cnt->conf.filepath, filename, imageext(cnt));
		sprintf(fullfilenamem, "%s/%s.%s", cnt->conf.filepath, filenamem, imageext(cnt));
	}
	if (conf->motion_img) {
		cnt->filesmc++;
		cnt->filesm=myrealloc(cnt->filesm, cnt->filesmc*(sizeof(void *)), "event_image_detect(1)");
		cnt->filesm[cnt->filesmc-1]=mymalloc(strlen(fullfilenamem)+1);
		strcpy(cnt->filesm[cnt->filesmc-1], fullfilenamem);
		cnt->filesmrate=myrealloc(cnt->filesmrate, cnt->filesmc*(sizeof(void *)), "event_image_detect(2)");
		cnt->filesmrate[cnt->filesmc-1]=cnt->lastrate;
		put_picture(cnt, fullfilenamem, cnt->imgs.out);
	
	}
	if (conf->new_img) {
		cnt->filesc++;
		cnt->files=myrealloc(cnt->files, cnt->filesc*(sizeof(void *)), "event_image_detect(3)");
		cnt->files[cnt->filesc-1]=mymalloc(strlen(fullfilename)+1);
		strcpy(cnt->files[cnt->filesc-1], fullfilename);
		cnt->filesrate=myrealloc(cnt->filesrate, cnt->filesc*(sizeof(void *)), "event_image_detect(4)");
		cnt->filesrate[cnt->filesc-1]=cnt->lastrate;
		put_picture(cnt, fullfilename, newimg);
	}
}

static void event_image_snapshot(int type, struct context *cnt, char *img, void *dummy2, struct tm *tm)
{
	char filepath[PATH_MAX];
	char filename[PATH_MAX];
	char fullfilename[PATH_MAX];
	char linkpath[PATH_MAX];

	sprintf(linkpath, "%s/lastsnap.%s", cnt->conf.filepath, imageext(cnt));
	remove(linkpath);
	if ( !strcmp(cnt->conf.snappath, "lastsnap") ) {
		char *snappath;
		/* conf.snappath would normally be defined but if someone deleted it by control interface
		   it is better to revert to the default than fail */
		if (cnt->conf.snappath)
			snappath = cnt->conf.snappath;
		else
			snappath = DEF_SNAPPATH;
			
		mystrftime(filepath, sizeof(filepath), snappath, cnt->currenttime, cnt->event_nr, cnt->shots);
		sprintf(filename, "%s.%s", filepath, imageext(cnt));
		sprintf(fullfilename, "%s/%s", cnt->conf.filepath, filename);
		symlink(filename, linkpath);
	} else {
		sprintf(fullfilename, "%s/lastsnap.%s", cnt->conf.filepath, imageext(cnt));
	}
	put_picture(cnt, fullfilename, img);
	cnt->snapshot=0;
}

#ifdef HAVE_FFMPEG
static void grey2yuv420p(unsigned char *u, unsigned char *v, int width, int height)
{
	memset(u, 128, width*height/4);
	memset(v, 128, width*height/4);
}
/*
void rgb2yuv420p(unsigned char *src, unsigned char *dst, int width, int height)
{
	unsigned char *y, *u, *v;
	unsigned char *r, *g, *b;
	int size, i=0;

	b=src;
	g=b+1;
	r=g+1;
	y=dst;
	u=y+width*height;
	v=u+(width*height)/4;
	memset(u, 0, width*height/4);
	memset(v, 0, width*height/4);

	size=width*height;
	while (i<size) {
		*y++=(9796**r+19235**g+3736**b)>>15;
		*u+=((-4784**r-9437**g+14221**b)>>17)+32;
		*v+=((20218**r-16941**g-3277**b)>>17)+32;
		r+=3;
		g+=3;
		b+=3;
		*y++=(9796**r+19235**g+3736**b)>>15;
		*u+=((-4784**r-9437**g+14221**b)>>17)+32;
		*v+=((20218**r-16941**g-3277**b)>>17)+32;
		r+=3;
		g+=3;
		b+=3;
		u++;
		v++;
		i+=2;
		if (0==i%width && 0!=i%(width*2)) {
			u-=width/2;
			v-=width/2;
		}
	}
}
*/

static void onffmpegclose_command(int type, struct context *cnt, char *filename, void *mime, struct tm *tm)
{
	if (!strncmp(mime, "video", 5))
		if (cnt->conf.onffmpegclosecommand) {
			exec_command(cnt, cnt->conf.onffmpegclosecommand, filename);
	}
}

static void event_ffmpeg_newfile(int type, struct context *cnt, char *img, void *dummy2, struct tm *currenttime)
{
	int width=cnt->imgs.width;
	int height=cnt->imgs.height;
	unsigned char *convbuf, *y, *u, *v;
	int fps;
	char stamp[PATH_MAX];
	char *mpegpath;

	if (!cnt->conf.ffmpeg_cap_new && !cnt->conf.ffmpeg_cap_motion)
		return;
		
	/* conf.mpegpath would normally be defined but if someone deleted it by control interface
	   it is better to revert to the default than fail */
	if (cnt->conf.mpegpath)
		mpegpath = cnt->conf.mpegpath;
	else
		mpegpath = DEF_MPEGPATH;
	mystrftime(stamp, sizeof(stamp), mpegpath, cnt->currenttime, cnt->event_nr, cnt->shots);
	/* motion mpegs get the same name as normal mpegs plus an appended 'm' */
	sprintf(cnt->motionfilename, "%s/%sm.mpg", cnt->conf.filepath, stamp);
	sprintf(cnt->newfilename, "%s/%s.mpg", cnt->conf.filepath, stamp);

	if (cnt->conf.ffmpeg_cap_new) {
		if (cnt->imgs.type==VIDEO_PALETTE_GREY) {
			convbuf=mymalloc((width*height)/2);
			y=img;
			u=convbuf;
			v=convbuf+(width*height)/4;
			grey2yuv420p(u, v, width, height);
		} else {
			convbuf=NULL;
			y=img;
			u=img+width*height;
			v=u+(width*height)/4;
		}
		if (cnt->conf.low_cpu)
			fps=cnt->conf.frame_limit;
		else
			fps=cnt->lastrate;
		if (fps>30)
			fps=30;
		if (fps<2)
			fps=2;
		cnt->ffmpeg_new=ffmpeg_open(cnt->newfilename, y, u, v, cnt->imgs.width, cnt->imgs.height, fps, cnt->conf.ffmpeg_bps);
		((struct ffmpeg *)cnt->ffmpeg_new)->udata=convbuf;
		event(EVENT_FILECREATE, cnt, cnt->newfilename, "video/mpeg", NULL);
	}
	if (cnt->conf.ffmpeg_cap_motion) {
		if (cnt->imgs.type==VIDEO_PALETTE_GREY) {
			convbuf=mymalloc((width*height)/2);
			y=cnt->imgs.out;
			u=convbuf;
			v=convbuf+(width*height)/4;
			grey2yuv420p(u, v, width, height);
		} else {
			y=cnt->imgs.out;
			u=cnt->imgs.out+width*height;
			v=u+(width*height)/4;
			convbuf=NULL;
		}
		if (cnt->conf.low_cpu)
			fps=cnt->conf.frame_limit;
		else
			fps=cnt->lastrate;
		if (fps>30)
			fps=30;
		if (fps<2)
			fps=2;
		cnt->ffmpeg_motion=ffmpeg_open(cnt->motionfilename, y, u, v, cnt->imgs.width, cnt->imgs.height, fps, cnt->conf.ffmpeg_bps);
		cnt->ffmpeg_motion->udata=convbuf;
		event(EVENT_FILECREATE, cnt, cnt->motionfilename, "video/mpeg", NULL);
	}
}

static void event_ffmpeg_timelaps(int type, struct context *cnt, char *img, void *dummy2, struct tm *currenttime)
{
	int width=cnt->imgs.width;
	int height=cnt->imgs.height;
	char timelapsfilename[PATH_MAX];
	unsigned char *convbuf, *y, *u, *v;

	if (!cnt->ffmpeg_timelaps) {
		char tmp[PATH_MAX], *timepath;

		/* conf.timepath would normally be defined but if someone deleted it by control interface
		   it is better to revert to the default than fail */
		if (cnt->conf.timepath)
			timepath = cnt->conf.timepath;
		else
			timepath = DEF_TIMEPATH;
		
		mystrftime(tmp, sizeof(tmp), timepath, currenttime, cnt->event_nr, cnt->shots);
		sprintf(timelapsfilename, "%s/%s.mpg", cnt->conf.filepath, tmp);
		if (cnt->imgs.type==VIDEO_PALETTE_GREY) {
			convbuf=mymalloc((width*height)/2);
			y=img;
			u=convbuf;
			v=convbuf+(width*height)/4;
			grey2yuv420p(u, v, width, height);
		} else {
			convbuf=NULL;
			y=img;
			u=img+width*height;
			v=u+(width*height)/4;
		}
		cnt->ffmpeg_timelaps=ffmpeg_open(timelapsfilename, y, u, v, cnt->imgs.width, cnt->imgs.height, 24, cnt->conf.ffmpeg_bps);
		cnt->ffmpeg_timelaps->udata=convbuf;
		event(EVENT_FILECREATE, cnt, timelapsfilename, "video/mpeg", NULL);
	}
	ffmpeg_put_image(cnt->ffmpeg_timelaps);
}

static void event_ffmpeg_put(int type, struct context *cnt, char *img, void *dummy2, struct tm *tm)
{
	if (cnt->ffmpeg_new)
	{
		if (cnt->imgs.type == VIDEO_PALETTE_GREY)
		{
			ffmpeg_put_image(cnt->ffmpeg_new);
		}
		else
		{
			int width=cnt->imgs.width;
			int height=cnt->imgs.height;
			unsigned char *y = img;
			unsigned char *u = y + (width * height);
			unsigned char *v = u + (width * height) / 4;

			ffmpeg_put_other_image(cnt->ffmpeg_new, y, u, v);
		}
	}
	if (cnt->ffmpeg_motion) {
		ffmpeg_put_image(cnt->ffmpeg_motion);
	}
}

static void event_ffmpeg_closefile(int type, struct context *cnt, char *dummy, void *dummy2, struct tm *tm)
{
	if (cnt->ffmpeg_new) {
		if (cnt->ffmpeg_new->udata)
			free(cnt->ffmpeg_new->udata);
		ffmpeg_close(cnt->ffmpeg_new);
		cnt->ffmpeg_new=NULL;

		event(EVENT_FILECLOSE, cnt, cnt->newfilename, "video/mpeg", NULL);
	}
	if (cnt->ffmpeg_motion) {
		if (cnt->ffmpeg_motion->udata)
			free(cnt->ffmpeg_motion->udata);
		ffmpeg_close(cnt->ffmpeg_motion);
		cnt->ffmpeg_motion=NULL;
	}
}

static void event_ffmpeg_timelapsend(int type, struct context *cnt, char *dummy, void *dummy2, struct tm *tm)
{
	if (cnt->ffmpeg_timelaps) {
		if (cnt->ffmpeg_timelaps->udata)
			free(cnt->ffmpeg_timelaps->udata);
		ffmpeg_close(cnt->ffmpeg_timelaps);
		cnt->ffmpeg_timelaps=NULL;
	}
}

#endif /* HAVE_FFMPEG */

static void event_predict_result(int type, struct context *cnt, char *dummy, void *event_tm, struct tm *currenttime)
{
	struct tm *eventtime=event_tm;
	char fullfilename[PATH_MAX];

	if (cnt->predict.length && cnt->conf.predict_enable) {
		char stamp[PATH_MAX], *predictpath;

		/* conf.predictpath would normally be defined but if someone deleted it by control interface
	   	   it is better to revert to the default than fail */
		if (cnt->conf.predictpath)
			predictpath = cnt->conf.predictpath;
		else
			predictpath = DEF_PREDICTPATH;
	
		mystrftime(stamp, sizeof(stamp), predictpath, eventtime, cnt->event_nr, cnt->shots);
		sprintf(fullfilename, "%s/%s.desc", cnt->conf.filepath, stamp);
		alg_predict_result(cnt, fullfilename);
		event(EVENT_FILECREATE, cnt, fullfilename, "text/plain", NULL);
	}
}

/*  
 *	Starting point for all events
 */

struct event_handlers {
	int type;
	event_handler handler;
};

struct event_handlers event_handlers[] = {
	{
	EVENT_FILECREATE,
	event_newfile
	},
	{
	EVENT_FILECREATE,
	onsave_command
	},
#ifdef HAVE_MYSQL
	{
	EVENT_FILECREATE,
	event_mysql_newfile
	},
#endif
#ifdef HAVE_PGSQL
	{
	EVENT_FILECREATE,
	event_pgsql_newfile
	},
#endif
	{
	EVENT_MOTION,
	event_motion
	},
	{
	EVENT_FIRSTMOTION,
	event_firstmotion
	},
	{
	EVENT_IMAGE_DETECTED,
	event_image_detect
	},
	{
	EVENT_IMAGE_SNAPSHOT,
	event_image_snapshot
	},
	{
	EVENT_IMAGE | EVENT_IMAGEM,
	event_vid_putpipe
	},
	{
	EVENT_IMAGE | EVENT_IMAGE_DETECTED,
	event_webcam_put
	},
#ifdef HAVE_FFMPEG
	{
	EVENT_FIRSTMOTION,
	event_ffmpeg_newfile
	},
	{
	EVENT_IMAGE_DETECTED,
	event_ffmpeg_put
	},
	{
	EVENT_ENDMOTION,
	event_ffmpeg_closefile
	},
	{
	EVENT_TIMELAPS,
	event_ffmpeg_timelaps
	},
	{
	EVENT_TIMELAPSEND,
	event_ffmpeg_timelapsend
	},
	{
	EVENT_FILECLOSE,
	onffmpegclose_command
	},
#endif
	{
	EVENT_ENDMOTION,
	event_predict_result
	},
	{
	EVENT_STOP,
	event_stop_webcam
	},
	{0, NULL}
};

void event (int type, struct context *cnt, char *filename, void *eventdata, struct tm *tm)
{
	int i=-1;

	while (event_handlers[++i].handler) {
		if (type & event_handlers[i].type)
			event_handlers[i].handler(type, cnt, filename, eventdata, tm);
	}
}
